/*
 * Copyright: (C) 1997, GNU (Robert Roebling)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef __GNUG__
#pragma implementation "WIN_dnd.h"
#endif

#define Uses_wxPanel
#define Uses_wxCanvas
#define Uses_XtIntrinsic
#include "wx.h"

#include <X11/Intrinsic.h>
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/Xmu/WinUtil.h>

#include "WIN_dnd.h"

//-----------------------------------------------------------------------------
//  cursor data
//-----------------------------------------------------------------------------

static Bool cursors_ready = FALSE;

#include "X11/cursorfont.h"

#include "cursor/file.xbm"
#include "cursor/file_mask.xbm"
#include "cursor/files.xbm"
#include "cursor/files_mask.xbm"
#include "cursor/dir.xbm"
#include "cursor/dir_mask.xbm"
#include "cursor/text.xbm"
#include "cursor/text_mask.xbm"
#include "cursor/grey.xbm"
#include "cursor/grey_mask.xbm"
#include "cursor/link.xbm"
#include "cursor/link_mask.xbm"
#include "cursor/app.xbm"
#include "cursor/app_mask.xbm"
#include "cursor/url.xbm"
#include "cursor/url_mask.xbm"
#include "cursor/mime.xbm"
#include "cursor/mime_mask.xbm"

typedef struct
{	
    int	Width,Height;
    unsigned char *ImageData,*MaskData;
    int	HotSpotX,HotSpotY;
    Pixmap	ImagePixmap,MaskPixmap;
    Cursor	cursorID;
} cursorData;

static cursorData dndCursor[DndEND]={
  { 0,0,NULL,NULL,0,0,0 },
  { grey_width,	grey_height,grey_bits,grey_mask_bits,
    grey_x_hot,grey_y_hot},
  { file_width,file_height,file_bits,file_mask_bits,
    file_x_hot,file_y_hot},
  { files_width,files_height,files_bits,files_mask_bits,
    files_x_hot,files_y_hot},
  { text_width,text_height,text_bits,text_mask_bits,
    text_x_hot,text_y_hot },
  { dir_width,dir_height,dir_bits,dir_mask_bits,
    dir_x_hot,dir_y_hot },
  { link_width,link_height,link_bits,link_mask_bits,
    link_x_hot,link_y_hot},
  { app_width,app_height,app_bits,app_mask_bits,
    app_x_hot,app_y_hot },
  { url_width,url_height,url_bits,url_mask_bits,
    url_x_hot,url_y_hot },
  { mime_width,mime_height,mime_bits,mime_mask_bits,
    mime_x_hot,mime_y_hot }  
};

//-----------------------------------------------------------------------------
//  wxDragCanvas
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxDragCanvas,wxCanvas);

wxDragCanvas::wxDragCanvas( wxWindow *parent, int x, int y, int width, int height, 
      long style, Constdata char *name ) : 
  wxCanvas( parent, x, y, width, height, style, name )
{
  isDragging = FALSE;
  dragCounter = 0;
  dragStartX = 0;
  dragStartY = 0;

  Display *display = wxGetDisplay();

  if (cursors_ready) return;

  cursors_ready = TRUE;

  int screen = DefaultScreen( display );
  Colormap colormap = DefaultColormap( display, screen );
  Window root = DefaultRootWindow( display );

  XColor black,white; // MARKUS: only locally needed
  black.pixel = BlackPixel( display, screen );
  white.pixel = WhitePixel( display, screen );
  XQueryColor( display, colormap, &black );
  XQueryColor( display, colormap, &white);

  for( int i = 1; i != DndEND; i++ )
  {
	dndCursor[i].ImagePixmap=
	    XCreateBitmapFromData( display, root,
				  (const char*)dndCursor[i].ImageData,
				  dndCursor[i].Width,
				  dndCursor[i].Height);
	dndCursor[i].MaskPixmap=
	    XCreateBitmapFromData( display, root,
				  (const char*)dndCursor[i].MaskData,
				  dndCursor[i].Width,
				  dndCursor[i].Height);
	dndCursor[i].cursorID=
	    XCreatePixmapCursor( display, dndCursor[i].ImagePixmap,
				dndCursor[i].MaskPixmap,
				&black,&white,
				dndCursor[i].HotSpotX,
				dndCursor[i].HotSpotY);
  };
	
  dndCursor[0].cursorID = XCreateFontCursor( display, XC_question_arrow );
  dndData = NULL;
  dndSize = 0;
  dndType = DndFiles;
};

void wxDragCanvas::OnEvent( wxMouseEvent &event )
{
  if (event.Dragging())
  {
    if (dragCounter < 5)
    { 
      dragCounter += 1;
      return;
    };
    
    float fx = 0;
    float fy = 0;
    event.Position( &fx, &fy );
    int x = (int)fx;
    int y = (int)fy;
    StartDragging( x, y );
    
    if (dndSize == 0)
    {
      dragCounter = 0;
      return;
    };  
    
    Display *display = wxGetDisplay();
    Window root = DefaultRootWindow( display );
    
    XChangeProperty( display, root, XInternAtom( display, "DndSelection", False) , XA_STRING, 8, PropModeReplace, 
      (const unsigned char*) dndData, dndSize ); 
    
    XEvent Event;
    XtAppContext app = XtWidgetToApplicationContext( HWidget() );
    XGrabPointer( display, root, False, ButtonMotionMask|ButtonPressMask|ButtonReleaseMask, 
      GrabModeSync, GrabModeAsync, root, dndCursor[dndType].cursorID, CurrentTime );
    int Dragging = 1; 
    int RootFlag = 0;
    while (Dragging)
    {
      XAllowEvents( display, SyncPointer, CurrentTime );
      XtAppNextEvent( app, &Event );
      switch (Event.type)
      {
        case ButtonRelease:
	  if (Event.xbutton.subwindow) RootFlag=0; else RootFlag=1;
	  Dragging=0;
	  break;
	default:
	  XtDispatchEvent( &Event );
	  break;
      };
    };
    XUngrabPointer( display, CurrentTime );
    
    Window dispatchWindow;
    Window target;
    if(!RootFlag)
    {
      target = XmuClientWindow( display, Event.xbutton.subwindow );
      if (target == Event.xbutton.subwindow)
         dispatchWindow = target;
      else 
         dispatchWindow = PointerWindow;
    }
    else
    {
      dispatchWindow = target = XtWindow( HWidget() );
    };
    int DropX = Event.xbutton.x_root;
    int DropY = Event.xbutton.y_root;
    Event.xclient.type			= ClientMessage;
    Event.xclient.display		= display;
    Event.xclient.message_type		= XInternAtom( display, "DndProtocol", False );
    Event.xclient.format		= 32;
    Event.xclient.window		= target;
    Event.xclient.data.l[0]		= dndType;
    Event.xclient.data.l[1]		= 0;
    Event.xclient.data.l[2]		= (long)XtWindow( HWidget() );
    Event.xclient.data.l[3]		= DropX;
    Event.xclient.data.l[4]		= DropY;
    XSendEvent( display, dispatchWindow, True, NoEventMask, &Event );
    XSync( display, FALSE );
    
    dragCounter = 0;
    delete [] dndData;
    dndData = NULL;
    dndSize = 0;
    return;
  };

  if (event.LeftUp() | event.RightUp())
  {
    dragCounter = 0;
    return;
  }; 
};

void wxDragCanvas::AddDragString( char *s )
{
  if (!s) return;
  int len = strlen(s);
  if (!len) return;
  int newSize = len+1 + dndSize;
  if (dndSize == 0) newSize++;
  char *newData = wxNEW char[newSize];
  if (dndSize > 0)
  {
    memcpy( newData, dndData, dndSize );
    memcpy( (char*)(&newData[dndSize-1]), s, len);
  }
  else
  {
    memcpy( (char*)(&newData[0]), s, len);
  };
  if (dndData) delete [] dndData;
  dndData = newData;
  dndData[newSize-1] = 0;
  dndData[newSize-2] = 0;
  dndSize = newSize;
};




