/*
 * Copyright: (C) 1997, GNU (Robert Roebling)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef __GNUG__
#pragma implementation "WIN_rawlistctrl.h"
#endif

#define Uses_wxPanel
#define Uses_wxCanvas
#define Uses_wxString
#define Uses_wxText
#define Uses_DialogBox
#include "wx.h"

#include "WIN_listctrl.h"
#include "WIN_rawlistctrl.h"

//-----------------------------------------------------------------------------
//  wxRawListItemInfo
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxRawListItemInfo,wxObject);

wxRawListItemInfo::wxRawListItemInfo(void)
{
  m_image = -1;
  m_data = 0;
  m_xpos = 0;
  m_ypos = 0;
  m_width = 0;
  m_height = 0;
  m_colour = wxBLACK;
};

wxRawListItemInfo::wxRawListItemInfo( const wxListItem &info )
{
  m_image = -1;
  m_data = 0;
  m_colour = info.m_colour;
  SetItem( info );
};

void wxRawListItemInfo::SetItem( const wxListItem &info )
{
  if (info.m_mask & wxLIST_MASK_TEXT) m_text = info.m_text;
  if (info.m_mask & wxLIST_MASK_IMAGE) m_image = info.m_image;
  if (info.m_mask & wxLIST_MASK_DATA) m_data = info.m_data;
  m_colour = info.m_colour;
  m_xpos = 0;
  m_ypos = 0;
  m_width = info.m_width;
  m_height = 0;
};

void wxRawListItemInfo::SetText( const wxString &s )
{
  m_text = s;
};

void wxRawListItemInfo::SetImage( const int image )
{
  m_image = image;
};

void wxRawListItemInfo::SetData( const long data )
{
  m_data = data;
};

void wxRawListItemInfo::SetPosition( const int x, const int y )
{
  m_xpos = x;
  m_ypos = y;
};

void wxRawListItemInfo::SetSize( int const width, const int height )
{
  m_width = width;
  m_height = height;
};

void wxRawListItemInfo::SetColour( wxColour *col )
{
  m_colour = col;
};

Bool wxRawListItemInfo::HasImage(void) const
{
  return (m_image >= 0);
};

Bool wxRawListItemInfo::HasText(void) const
{
  return (!m_text.IsNull());
};

Bool wxRawListItemInfo::IsHit( const int x, const int y ) const
{
  return ((x >= m_xpos) && (x <= m_xpos+m_width) && (y >= m_ypos) && (y <= m_ypos+m_height));
};

void wxRawListItemInfo::GetText( wxString &s )
{
  s = m_text;
};

int wxRawListItemInfo::GetX( void ) const
{
  return m_xpos;
};

int wxRawListItemInfo::GetY( void ) const
{
  return m_ypos;
};

int wxRawListItemInfo::GetWidth(void) const
{
  return m_width;
};

int wxRawListItemInfo::GetHeight(void) const
{
  return m_height;
};

int wxRawListItemInfo::GetImage(void) const
{
  return m_image;
};

void wxRawListItemInfo::GetItem( wxListItem &info )
{
  info.m_text = m_text;
  info.m_image = m_image;
  info.m_data = m_data;
};

wxColour *wxRawListItemInfo::GetColour(void)
{
  return m_colour;
};

//-----------------------------------------------------------------------------
//  wxRawListColumnInfo
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxRawListColumnInfo,wxObject);

wxRawListColumnInfo::wxRawListColumnInfo(void)
{
  m_mask = 0;
  m_image = 0;
  m_format = 0;
  m_width = 0;
  m_xpos = 0;
  m_ypos = 0;
  m_height = 0;
};

wxRawListColumnInfo::wxRawListColumnInfo( const wxListItem &item )
{
  SetItem( item );
  m_xpos = 0;
  m_ypos = 0;
  m_height = 0;
};

void wxRawListColumnInfo::SetItem( const wxListItem &item )
{
  m_mask = item.m_mask;
  m_text = item.m_text;
  m_image = item.m_image;
  m_format = item.m_format;
  m_width = item.m_width;
  if (m_width < 0) m_width = 80;
  if (m_width < 6) m_width = 6;
};

void wxRawListColumnInfo::SetPosition( const int x, const int y )
{
  m_xpos = x;
  m_ypos = y;
};

void wxRawListColumnInfo::SetHeight( const int h )
{
  m_height = h;
};

void wxRawListColumnInfo::SetWidth( const int w )
{
  m_width = w;
  if (m_width < 0) m_width = 80;
  if (m_width < 6) m_width = 6;
};

void wxRawListColumnInfo::SetFormat( const int format )
{
  m_format = format;
};

Bool wxRawListColumnInfo::HasImage(void) const
{
  return (m_image != 0);
};

Bool wxRawListColumnInfo::HasText(void) const
{
  return (m_text.Length() > 0);
};

Bool wxRawListColumnInfo::IsHit( int x, int y ) const
{
  return ((x >= m_xpos) && (x <= m_xpos+m_width) && (y >= m_ypos) && (y <= m_ypos+m_height));
};

void wxRawListColumnInfo::GetItem( wxListItem &item )
{
  item.m_mask = m_mask;
  item.m_text = m_text;
  item.m_image = m_image;
  item.m_format = m_format;
  item.m_width = m_width;
};

void wxRawListColumnInfo::GetText( wxString &s )
{
  s =  m_text;
};

int wxRawListColumnInfo::GetImage(void) const
{
  return m_image;
};

int wxRawListColumnInfo::GetWidth(void) const
{
  return m_width;
};

int wxRawListColumnInfo::GetFormat(void) const
{
  return m_format;
};

//-----------------------------------------------------------------------------
//  wxRawListLine
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxRawListLine,wxObject);

wxRawListLine::wxRawListLine( wxRawListCtrl *owner, const int mode, wxBrush *hilightBrush )
{
  m_mode = mode;
  m_hilighted = FALSE;
  m_owner = owner;
  m_hilightBrush = hilightBrush;
  m_items.DeleteContents( TRUE );
  m_spacing = 0;
};

void wxRawListLine::CalculateSize( wxCanvasDC *dc, const int spacing )
{
  m_spacing = spacing;
  switch (m_mode)
  {
    case wxLC_ICON:
    {
      m_bound_all.width = m_spacing;
      m_bound_all.height = m_spacing+13;
      wxNode *node = m_items.First();
      if (node)
      {
        wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
        wxString s;
        item->GetText( s );
        float fw,fh;
        dc->GetTextExtent( s, &fw, &fh );
	if ((int)fw > m_spacing) m_bound_all.width = (int)fw;
      };
      break;
    };
    case wxLC_LIST:
    {
      wxNode *node = m_items.First();
      if (node)
      {
        wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
        wxString s;
        item->GetText( s );
        float fw,fh;
        dc->GetTextExtent( s, &fw, &fh );
        m_bound_all.width = (int)fw;
        m_bound_all.height = (int)fh;
      };
      break;
    };
    case wxLC_REPORT:
    {
      m_bound_all.width = 0;
      m_bound_all.height = 0;
      wxNode *node = m_items.First();
      while (node)
      {
        wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
        wxString s;
        item->GetText( s );
	if (s.IsNull()) s = "H";
        float fw,fh;
        dc->GetTextExtent( s, &fw, &fh );
	int w = (int)fw;
	int h = (int)fh;
	item->SetSize( item->GetWidth(), h );
        m_bound_all.width += w;
        m_bound_all.height = h;
	node = node->Next();
      };
      break;
    };
  };
};

void wxRawListLine::SetPosition( wxCanvasDC *dc, const int x, const int y, const int window_width )
{
  m_bound_all.x = x;
  m_bound_all.y = y;
  switch (m_mode)
  {
    case wxLC_ICON:
    {
      AssignRect( m_bound_icon, 0, 0, 0, 0 );
      AssignRect( m_bound_label, 0, 0, 0, 0 );
      AssignRect( m_bound_hilight, m_bound_all );
      wxNode *node = m_items.First();
      if (node)
      {
        wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
	if (item->HasImage())
	{
          wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
	  int w = 0;
	  int h = 0;
	  m_owner->GetImageSize( item->GetImage(), w, h );
	  m_bound_icon.x = m_bound_all.x + (m_spacing/2) - (w/2);
	  m_bound_icon.y = m_bound_all.y + m_spacing - h - 5;
	  m_bound_icon.width = w;
	  m_bound_icon.height = h;
	  if (!item->HasText())
	  {
	    AssignRect( m_bound_hilight, m_bound_icon );
	    m_bound_hilight.x -= 3;
	    m_bound_hilight.y -= 3;
	    m_bound_hilight.width += 7;
	    m_bound_hilight.height += 7;
	  };
	};
	if (item->HasText())
	{
          wxString s;
          item->GetText( s );
          float fw,fh;
          dc->GetTextExtent( s, &fw, &fh );
	  if (m_bound_all.width > m_spacing)
	    m_bound_label.x = m_bound_all.x;
	  else
	    m_bound_label.x = m_bound_all.x +  (m_spacing/2) - (int)(fw/2);
	  m_bound_label.y = m_bound_all.y + m_bound_all.height - (int)(fh);
          m_bound_label.width = (int)fw;
          m_bound_label.height = (int)fh;
	  AssignRect( m_bound_hilight, m_bound_label );
	};
      };
      break;
    };
    case wxLC_LIST:
    {
      AssignRect( m_bound_label, m_bound_all );
      AssignRect( m_bound_hilight, m_bound_all );
      AssignRect( m_bound_icon, 0, 0, 0, 0 );
      break;
    };
    case wxLC_REPORT:
    {
      float fw,fh;
      dc->GetTextExtent( "H", &fw, &fh );
      m_bound_all.height = (int)fh;
      m_bound_all.width = window_width;
      AssignRect( m_bound_label, m_bound_all );
      AssignRect( m_bound_hilight, m_bound_all );
      AssignRect( m_bound_icon, 0, 0, 0, 0 );
      m_bound_hilight.width = window_width-7;
      m_bound_label.width = window_width-7;
      break;
    };
  };
};

void wxRawListLine::SetColumnPosition( const int index, const int x )
{
  int i = index;
  wxNode *node = m_items.Nth( i );
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    item->SetPosition( x, m_bound_all.y );
  };  
};

void wxRawListLine::GetSize( int &width, int &height )
{
  width = m_bound_all.width;
  height = m_bound_all.height;
};

void wxRawListLine::GetExtent( int &x, int &y, int &width, int &height )
{
  x = m_bound_all.x;
  y = m_bound_all.y;
  width = m_bound_all.width;
  height = m_bound_all.height;
};

void wxRawListLine::GetRect( wxRectangle &rect )
{
  AssignRect( rect, m_bound_all );
};

long wxRawListLine::IsHit( const int x, const int y )
{
  wxNode *node = m_items.First();
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    if (item->HasImage() && IsInRect( x, y, m_bound_icon )) return wxLIST_HITTEST_ONITEMICON;
    if (item->HasText() && IsInRect( x, y, m_bound_label )) return wxLIST_HITTEST_ONITEMLABEL;
    if (!(item->HasImage() || item->HasText())) return 0;
  };
  // if there is no icon or text = empty
  if (IsInRect( x, y, m_bound_all )) return wxLIST_HITTEST_ONITEMICON;
  return 0;
};

void wxRawListLine::InitItems( const int num )
{
  for (int i = 0; i < num; i++) m_items.Append( new wxRawListItemInfo() );
};

void wxRawListLine::SetItem( const int index, const wxListItem &info )
{
  wxNode *node = m_items.Nth( index );
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    item->SetItem( info );
  };
};

void wxRawListLine::GetItem( int const index, wxListItem &info )
{
  int i = index;
  wxNode *node = m_items.Nth( i );
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    item->GetItem( info );
  };
};

void wxRawListLine::GetText( const int index, wxString &s )
{
  int i = index;
  wxNode *node = m_items.Nth( i );
  s = "";
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    item->GetText( s );
  };
};

void wxRawListLine::SetText( const int index, const wxString s )
{
  int i = index;
  wxNode *node = m_items.Nth( i );
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    item->SetText( s );
  };
};

int wxRawListLine::GetImage( const int index )
{
  int i = index;
  wxNode *node = m_items.Nth( i );
  if (node)
  {
    wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
    return item->GetImage();
  };
  return -1;
};

void wxRawListLine::DoDraw( wxCanvasDC *dc, const Bool hilight, const Bool paintBG )
{
  if (!m_owner->IsExposed( m_bound_all.x-2, m_bound_all.y-2, 
                           m_bound_all.width+4, m_bound_all.height+4 ) ) return;
  if (paintBG)
  {
    if (hilight)
    {
      dc->SetBrush( m_hilightBrush );
      dc->SetPen( wxTRANSPARENT_PEN );
    }
    else
    {
      dc->SetBrush( wxWHITE_BRUSH );
      dc->SetPen( wxTRANSPARENT_PEN );
    };
    dc->DrawRectangle( m_bound_hilight.x-2, m_bound_hilight.y-1,
                       m_bound_hilight.width+3, m_bound_hilight.height+1 );
  };
  if (m_mode == wxLC_REPORT)
  {
    wxString s;
    wxNode *node = m_items.First();
    while (node)
    {
      wxRawListItemInfo *info = (wxRawListItemInfo*)node->Data();
      dc->SetClippingRegion( info->GetX(), info->GetY(), info->GetWidth(), info->GetHeight() );
      info->GetText( s );
      dc->SetTextForeground( info->GetColour() );
      dc->DrawText( s, info->GetX()+2, info->GetY() );
      node = node->Next();
    };
    dc->DestroyClippingRegion();
  }
  else
  {
    wxNode *node = m_items.First();
    if (node)
    {
      wxRawListItemInfo *item = (wxRawListItemInfo*)node->Data();
      if (item->HasImage())
      {
        m_owner->DrawImage( item->GetImage(), dc, m_bound_icon.x, m_bound_icon.y );
      };
      if (item->HasText())
      {
        wxString s;
        item->GetText( s );
        dc->SetTextForeground( item->GetColour() );
        dc->DrawText( s, m_bound_label.x, m_bound_label.y );
      };
    };
  };
};

void wxRawListLine::DrawHilight( wxCanvasDC *dc, const Bool on )
{
  if (on == m_hilighted) return;
  if (on) 
    m_owner->SelectLine( this );
  else
    m_owner->DeselectLine( this );
  DoDraw( dc, on, TRUE );
  m_hilighted = on;
};

void wxRawListLine::ReverseHilight( wxCanvasDC *dc )
{
  m_hilighted = !m_hilighted;
  if (m_hilighted) 
    m_owner->SelectLine( this );
  else
    m_owner->DeselectLine( this );
  DoDraw( dc, m_hilighted, TRUE );
};

void wxRawListLine::DrawRubberBand( wxCanvasDC *dc, const Bool on )
{
  if (on)
  {
    dc->SetPen( wxBLACK_DASHED_PEN );
  }
  else
  {
    dc->SetPen( wxBLACK_DASHED_PEN );
    dc->SetPen( wxWHITE_PEN );
  };
  dc->SetBrush( wxTRANSPARENT_BRUSH );
  dc->DrawRectangle( m_bound_hilight.x-2, m_bound_hilight.y-2,
                     m_bound_hilight.width+3, m_bound_hilight.height+2 );
};

void wxRawListLine::Draw( wxCanvasDC *dc )
{
  DoDraw( dc, m_hilighted, m_hilighted );
};

Bool wxRawListLine::IsInRect( const int x, const int y, const wxRectangle &rect )
{
  return ((x >= rect.x) && (x <= rect.x+rect.width) && (y >= rect.y) && (y <= rect.y+rect.height));
};

Bool wxRawListLine::IsHilighted( void )
{
  return m_hilighted;
};

void wxRawListLine::AssignRect( wxRectangle &dest, const int x, const int y, const int width, const int height )
{
  dest.x = x;
  dest.y = y;
  dest.width = width;
  dest.height = height;
};

void wxRawListLine::AssignRect( wxRectangle &dest, const wxRectangle &source )
{
  dest.x = source.x;
  dest.y = source.y;
  dest.width = source.width;
  dest.height = source.height;
};

//-----------------------------------------------------------------------------
//  wxColumnTitles
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxColumnTitles,wxPanel);

wxColumnTitles::wxColumnTitles( wxWindow *win, wxRawListCtrl *owner, int x, int y, int width, int height, 
      long style, Constdata char *name ) :
      wxPanel( win, x, y, width, height, style | wxNO_KEYB_TRAVERSAL, name )
{
  m_owner = owner;
  m_currentCursor = wxSTANDARD_CURSOR;
  m_resizeCursor = wxNEW wxCursor( wxCURSOR_SIZEWE );
};

void wxColumnTitles::DoDrawRect( wxCanvasDC *dc, int x, int y, int w, int h )
{
  const m_corner = 5;

  dc->SetBrush( wxTRANSPARENT_BRUSH );

//  dc->SetPen( wxMEDIUM_GREY_PEN );
  dc->DrawLine( x+w-m_corner+1, y, x+w, y+h );  // right
 (outer)
  dc->DrawLine( x+w-m_corner, y, x+w-1, y+h );  // right
 (inner)
  dc->DrawRectangle( x+1, y+h-1, w-2, 0 );      // bottom
 (inner)
  dc->DrawRectangle( x, y+h, w, 0 );            // bottom
 (outer)
  
  dc->SetPen( wxWHITE_PEN );
  dc->DrawRectangle( x, y, w-m_corner+1, 0 );   // top
  (outer)
  dc->DrawRectangle( x, y+1, w-m_corner, 0 );   // top
  (inner)
  dc->DrawRectangle( x, y, 0, h );              // left
 (outer)
  dc->DrawRectangle( x+1, y, 0, h-1 );          // left
 (inner)
};

void wxColumnTitles::DoDrawFrame( wxCanvasDC *dc, int x, int y, int w, int h )
{
  dc->SetBrush( wxTRANSPARENT_BRUSH );

//  dc->SetPen( wxMEDIUM_GREY_PEN );
  dc->DrawRectangle( x, y, w, 1 );          // top
  (2 pixels)
  dc->DrawRectangle( x, y, 1, h );          // left
 (2 pixels)

  dc->SetPen( wxWHITE_PEN );
  dc->DrawRectangle( x+w, y, 0, h );        // right
 (outer)
  dc->DrawRectangle( x, y+h, w, 0 );        // bottom
 (outer)
  dc->DrawRectangle( x+w-1, y+1, 0, h-1 );  // right
 (inner)
  dc->DrawRectangle( x+1, y+h-1, w-3, 0 );  // bottom
 (inner)
};

void wxColumnTitles::OnPaint( void )
{
  wxCanvasDC *dc = GetDC();
  wxPen *oldPen = dc->GetPen();
  wxBrush *oldBrush = dc->GetBrush();
  wxFont* oldFont = dc->GetFont();
  dc->BeginDrawing();

  dc->SetFont( m_owner->GetMyFont() );

  int w = 0;
  int h = 0;
  int x = 0;
  int y = 0;
  GetClientSize( &w, &h );
  h -= 3;
  w -= 1;

  wxColour *myGray = new wxColour( 130, 130, 130 );  
  wxPen *myPen = wxThePenList->FindOrCreatePen( myGray, 1, wxSOLID );
  delete myGray;
  dc->SetPen( myPen );
  
  DoDrawFrame( dc, x, y, w, h );

  dc->SetTextForeground( wxBLACK );

  x = 2;
  y = 2;
  int numColumns = m_owner->RawGetColumnCount();
  wxListItem item;
  for (int i = 0; i < numColumns; i++)
  {
    m_owner->RawGetColumn( i, item );
    int cw = item.m_width-2;
    if ((i+1 == numColumns) || (x+item.m_width > w-5)) cw = w-x-4;
    dc->SetPen( myPen );
    
    DoDrawRect( dc, x, y, cw, h-4 );
    dc->SetClippingRegion( x, y, cw-5, h-4 );
    dc->DrawText( item.m_text, x+4, y+3 );
    dc->DestroyClippingRegion();
    x += item.m_width;
    if (x > w+5) break;
  }; 
  dc->EndDrawing();
  dc->SetPen( oldPen );
  dc->SetBrush( oldBrush );
  dc->SetFont( oldFont );
  
};

void wxColumnTitles::OnEvent( wxMouseEvent &event )
{
  float fx = 0;
  float fy = 0;
  event.Position( &fx, &fy );
  int x = (int)fx;
  int y = (int)fy;
  if (event.Moving())
  {
    Bool hit = FALSE;
    int xpos = 0;
    for (int j = 0; j < m_owner->RawGetColumnCount(); j++)
    {
      xpos += m_owner->RawGetColumnWidth( j );
      if ((abs(x-xpos) < 2) && (y < 14)) { hit = TRUE; break; }
    };
    if (hit)
    {
      if (m_currentCursor == wxSTANDARD_CURSOR) SetCursor( m_resizeCursor );
      m_currentCursor = m_resizeCursor;
    }
    else
    {
      if (m_currentCursor != wxSTANDARD_CURSOR) SetCursor( wxSTANDARD_CURSOR );
      m_currentCursor = wxSTANDARD_CURSOR;
    };
  }; 
};

void wxColumnTitles::OnSetFocus(void)
{
/*
  printf( "Focus.\n" );	
  m_owner->SetFocus();
*/
};


//-----------------------------------------------------------------------------
//  wxRawListCtrl
//-----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(wxRawListCtrl,wxDragCanvas);

    
wxRawListCtrl::wxRawListCtrl( void ) :
  wxDragCanvas()
{
  m_mode = 0;
  m_lines.DeleteContents( TRUE );
  m_columns.DeleteContents( TRUE );
  m_current = NULL;
  m_visibleLines = 0;
  m_hilightBrush = NULL;
  m_myFont = NULL;
  m_titlesWindow = NULL;
  m_xScroll = 0;
  m_yScroll = 0;
  m_dirty = FALSE;
  m_small_image_list = NULL;
  m_normal_image_list = NULL;
  m_small_spacing = 30;
  m_normal_spacing = 40;
};

wxRawListCtrl::wxRawListCtrl( wxWindow *parent, int x, int y, int width, int height, long mode, 
  long style, Constdata char *name ) :
  wxDragCanvas( parent, x, y, width, height, style, name )   
{
  m_mode = mode;
  m_lines.DeleteContents( TRUE );
  m_columns.DeleteContents( TRUE );
  m_current = NULL;
  m_visibleLines = 0;
  wxColour *m_hilightColour = new wxColour( 208, 180, 176 );
  m_hilightBrush = new wxBrush( (*m_hilightColour), wxSOLID );
  m_small_image_list = NULL;
  m_normal_image_list = NULL;
  m_small_spacing = 30;
  m_normal_spacing = 40;
  AllowDoubleClick( TRUE );
  m_myFont = ((wxPanel*)GetParent())->GetButtonFont();
  m_titlesWindow = new wxColumnTitles( parent, this, x+3, y+2, width-4, 13 );
  if (m_mode & wxLC_LIST)
  {
    m_xScroll = 15;
    m_yScroll = 0;
    SetScrollbars( m_xScroll, m_yScroll, 1, 0, 20, 0 );
  }
  else
  if (m_mode & wxLC_REPORT)
  {
    m_xScroll = 0;
    m_yScroll = 15;
    SetScrollbars( m_xScroll, m_yScroll, 1, 0, 0, 20 );
  }
  else
  {
    m_xScroll = 0;
    m_yScroll = 0;
  };
};

wxRawListCtrl::~wxRawListCtrl( void )
{
//  if (m_hilightColour) delete m_hilightColour;
//  if (m_hilightBrush) delete m_hilightBrush;
//  if (m_myFont) delete m_myFont;
  if (m_titlesWindow) delete m_titlesWindow;
};

void wxRawListCtrl::OnPaint( void )
{
  if (m_dirty) CalculatePositions();
  wxCanvasDC *dc = GetDC();
  wxPen *oldPen = dc->GetPen();
  wxBrush *oldBrush = dc->GetBrush();
  wxFont* oldFont = dc->GetFont();
  dc->BeginDrawing();

  dc->SetFont( m_myFont );

  wxNode *node = m_lines.First();
  while (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    line->Draw( dc );
    node = node->Next();
  };
  if (m_current) m_current->DrawRubberBand( dc, TRUE );

  dc->EndDrawing();
  dc->SetPen( oldPen );
  dc->SetBrush( oldBrush );
  dc->SetFont( oldFont );
};

void wxRawListCtrl::CalculatePositions( void )
{
  m_dirty = FALSE;
  
  wxCanvasDC *dc = GetDC();
  wxPen *oldPen = dc->GetPen();
  wxBrush *oldBrush = dc->GetBrush();
  wxFont* oldFont = dc->GetFont();
  dc->SetFont( m_myFont );

  int clientWidth = 0;
  int clientHeight = 0;
  GetClientSize( &clientWidth, &clientHeight );
  int iconSpacing = 0;
  if (m_mode & wxLC_ICON) iconSpacing = m_normal_spacing;
  if (m_mode & wxLC_SMALL_ICON) iconSpacing = m_small_spacing;
  wxNode *node = m_lines.First();
  while (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    line->CalculateSize( dc, iconSpacing );
    node = node->Next();
  };

  int x = 5;
  int y = 6;
  int lineWidth = 0;
  int lineHeight = 0;
  int lineSpacing = 0;  

  node = m_lines.First();
  if (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    int dummy = 0;
    line->GetSize( dummy, lineSpacing );
    lineSpacing += 6;
  }
  else
  {
    // just in case
    float fh = 0;
    float fw = 0;
    dc->GetTextExtent( "H", &fw, &fh );
    lineSpacing = 6 + (int)fh;  
  };
  
  
  if (m_mode & wxLC_REPORT)
  {
    int entireHeight = 0;
    node = m_lines.First();
    while (node)
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      line->SetPosition( dc, x, y, clientWidth );
      int col_x = 3;
      for (int i = 0; i < RawGetColumnCount(); i++)
      {
        line->SetColumnPosition( i, col_x );
        col_x += RawGetColumnWidth( i );
      };
      line->GetSize( lineWidth, lineHeight );
      entireHeight += lineHeight+6;
      y += lineSpacing;
      node = node->Next();
    };
    SetScrollRange( wxVERTICAL, entireHeight+22 );
  }
  else
  {
    int maxWidth = 0;
    int entireWidth = 0;
    node = m_lines.First();
    while (node)
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      line->SetPosition( dc, x, y, clientWidth );
      line->GetSize( lineWidth, lineHeight );
      if (lineWidth > maxWidth) maxWidth = lineWidth;
      y += lineSpacing;
      if (y+lineHeight > clientHeight) 
      { 
        y = 6;
        x += maxWidth+10; 
        entireWidth += maxWidth+10;
        maxWidth = 0;
      };
      node = node->Next();
    };
    entireWidth += maxWidth;
    SetScrollRange( wxHORIZONTAL, entireWidth+22  );
  };

  m_visibleLines = (clientHeight) / (lineSpacing);
  dc->SetPen( oldPen );
  dc->SetBrush( oldBrush );
  dc->SetFont( oldFont );
};

void wxRawListCtrl::HilightAll( const Bool on )
{
  wxCanvasDC *dc = GetDC();
  wxBrush *oldBrush = dc->GetBrush();
  wxPen *oldPen = dc->GetPen();
  dc->SetFont( m_myFont );
  wxNode *node = m_lines.First();
  while (node)
  {
    wxRawListLine *line = (wxRawListLine *)node->Data();
    line->DrawHilight( dc, on );
    node = node->Next();
  };
  dc->SetBrush( oldBrush );
  dc->SetPen( oldPen );
};

void wxRawListCtrl::ActivateLine( wxRawListLine *line )
{
  wxListEvent le( wxEVT_COMMAND_LIST_KEY_DOWN );
  le.m_code = 0;
  le.m_itemIndex = GetIndexOfLine( line );
  le.m_col = 0;
  line->GetItem( 0, le.m_item );
  OnListNotify( le );
};

void wxRawListCtrl::SendNotify( wxRawListLine *line, long command )
{
  wxListEvent le( command );
  le.m_code = 0;
  le.m_itemIndex = GetIndexOfLine( line );
  le.m_col = 0;
  line->GetItem( 0, le.m_item );
  OnListNotify( le );
};

void wxRawListCtrl::FocusLine( wxRawListLine *line )
{
  SendNotify( line, wxEVT_COMMAND_LIST_ITEM_FOCUSSED );
};

void wxRawListCtrl::UnfocusLine( wxRawListLine *line )
{
  SendNotify( line, wxEVT_COMMAND_LIST_ITEM_UNFOCUSSED );
};

void wxRawListCtrl::SelectLine( wxRawListLine *line )
{
  SendNotify( line, wxEVT_COMMAND_LIST_ITEM_SELECTED );
};

void wxRawListCtrl::DeselectLine( wxRawListLine *line )
{
  SendNotify( line, wxEVT_COMMAND_LIST_ITEM_DESELECTED );
};

void wxRawListCtrl::DeleteLine( wxRawListLine *line )
{
  SendNotify( line, wxEVT_COMMAND_LIST_DELETE_ITEM );
};

void wxRawListCtrl::RenameLine( wxRawListLine *line, const wxString &newName )
{
  wxListEvent le( wxEVT_COMMAND_LIST_END_LABEL_EDIT );
  le.m_code = 0;
  le.m_itemIndex = GetIndexOfLine( line );
  le.m_col = 0;
  line->GetItem( 0, le.m_item );
  le.m_item.m_text = newName;
  OnListNotify( le );
};

void wxRawListCtrl::OnEvent( wxMouseEvent &event )
{
  if (!m_current) return;
  if (m_dirty) return;
  wxDragCanvas::OnEvent( event );
  float fx = 0;
  float fy = 0;
  event.Position( &fx, &fy );
  int x = (int)fx;
  int y = (int)fy;
  if (event.ButtonDClick())
  {
    wxNode *node = m_lines.First();
    while (node)
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      if (line->IsHit( x, y ))
      {
        ActivateLine( line );
        return;
      };
      node = node->Next();
    };
    return;
  };
  if (event.LeftDown())
  {
    wxRawListLine *line = NULL;
    long hitResult = 0;
    wxNode *node = m_lines.First();
    while (node)
    {
       line = (wxRawListLine*)node->Data();
       hitResult = line->IsHit( x, y );
       if (hitResult) break;
       line = NULL;
       node = node->Next();
    };
    if (!line) return;
    
    if (!event.ShiftDown() || (m_mode & wxLC_SINGLE_SEL)) HilightAll( FALSE );

    wxCanvasDC *dc = GetDC();
    wxBrush *oldBrush = dc->GetBrush();
    wxPen *oldPen = dc->GetPen();
    dc->SetFont( m_myFont );

    m_current->DrawRubberBand( dc, FALSE );
    wxRawListLine *oldCurrent = m_current;

    if (line) m_current = line;
    m_current->ReverseHilight( dc );
    m_current->DrawRubberBand( dc, TRUE );
    if (m_current != oldCurrent)
    {
      UnfocusLine( oldCurrent );
      FocusLine( m_current );
    }
    else
    {
      if ((hitResult == wxLIST_HITTEST_ONITEMLABEL) && (m_mode & wxLC_ICON))
      {
        wxString s;
	m_current->GetText( 0, s );
	char *res = wxGetTextFromUser( "New label:", "User input", s );
	if (res)
	{
	  m_dirty = TRUE;
	  s = res;
	  RenameLine( m_current, s );
	};
      };
    };
    dc->SetBrush( oldBrush );
    dc->SetPen( oldPen );
    return;
  };
};

void wxRawListCtrl::MoveToFocus( void )
{
  if (!m_current) return;
  int x = 0;
  int y = 0;
  int w = 0;
  int h = 0;
  m_current->GetExtent( x, y, w, h );
  int w_p = 0;
  int h_p = 0;
  GetClientSize( &w_p, &h_p );
  if (m_mode & wxLC_REPORT)
  {
    if (GetScrollPos( wxHORIZONTAL ) != 0) SetScrollPos( wxHORIZONTAL, 0);
    int y_s = m_yScroll*GetScrollPos( wxVERTICAL );
    if ((y > y_s) && (y+h < y_s+h_p)) return; 
    if (y-y_s < 5) SetScrollPos( wxVERTICAL, (y-5)/m_yScroll );
    if (y+h+5 > y_s+h_p) SetScrollPos( wxVERTICAL, (y+h-h_p+h+5)/m_yScroll );
  }
  else
  {
    if (GetScrollPos( wxVERTICAL ) != 0) SetScrollPos( wxVERTICAL, 0);
    int x_s = m_xScroll*GetScrollPos( wxHORIZONTAL );
    if ((x > x_s) && (x+w < x_s+w_p)) return;
    if (x-x_s < 5) SetScrollPos( wxHORIZONTAL, (x-5)/m_xScroll );
    if (x+w > x_s+w_p)  SetScrollPos( wxHORIZONTAL, (x+w-w_p+5)/m_xScroll );
  };
};

void wxRawListCtrl::OnArrowChar( wxRawListLine *newCurrent, Bool shiftDown )
{
  m_current->DrawRubberBand( dc, FALSE );
  if (m_mode & wxLC_SINGLE_SEL) m_current->DrawHilight( dc, FALSE );
  UnfocusLine( m_current );
  m_current = newCurrent;
  MoveToFocus();
  if (shiftDown || (m_mode & wxLC_SINGLE_SEL)) m_current->DrawHilight( dc, TRUE );
  m_current->DrawRubberBand( dc, TRUE );
  FocusLine( m_current );
};

void wxRawListCtrl::OnChar( wxKeyEvent &event )
{
  if (event.KeyCode() == WXK_TAB)
  {
    if (event.ShiftDown())
      TravPrev( &event );
    else
      TravNext( &event );
    return;
  };
  if (!m_current) return;
  wxCanvasDC *dc = GetDC();
  wxBrush *oldBrush = dc->GetBrush();
  wxPen *oldPen = dc->GetPen();
  dc->SetFont( m_myFont );
  switch (event.KeyCode())
  {
    case WXK_UP:
    {
      wxNode *node = m_lines.Member( m_current )->Previous();
      if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_DOWN:
    {
      wxNode *node = m_lines.Member( m_current )->Next();
      if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_END:
    {
      wxNode *node = m_lines.Last();
      OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_HOME:
    {
      wxNode *node = m_lines.First();
      OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_PRIOR:
    {
      int steps = 0;
      if (m_mode & wxLC_REPORT) { steps = m_visibleLines-1; }
      else
      {
        int pos = 0; 
	wxNode *node = m_lines.First();
        for (;;) { if (m_current == (wxRawListLine*)node->Data()) break; pos++; node = node->Next(); };
        steps = pos % m_visibleLines;
      };
      wxNode *node = m_lines.Member( m_current );
      for (int i = 0; i < steps; i++) if (node->Previous()) node = node->Previous();
      if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_NEXT:
    {
      int steps = 0;
      if (m_mode & wxLC_REPORT) { steps = m_visibleLines-1; }
      else
      {
        int pos = 0; wxNode *node = m_lines.First();
        for (;;) { if (m_current == (wxRawListLine*)node->Data()) break; pos++; node = node->Next(); };
        steps = m_visibleLines-(pos % m_visibleLines)-1;
      };
      wxNode *node = m_lines.Member( m_current );
      for (int i = 0; i < steps; i++) if (node->Next()) node = node->Next();
      if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      break;
    };
    case WXK_LEFT:
    {
      if (!(m_mode & wxLC_REPORT))
      {
        wxNode *node = m_lines.Member( m_current );
        for (int i = 0; i <m_visibleLines; i++) if (node->Previous()) node = node->Previous();
        if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      };
      break;
    };
    case WXK_RIGHT:
    {
      if (!(m_mode & wxLC_REPORT))
      {
        wxNode *node = m_lines.Member( m_current );
        for (int i = 0; i <m_visibleLines; i++) if (node->Next()) node = node->Next();
        if (node) OnArrowChar( (wxRawListLine*)node->Data(), event.ShiftDown() );
      };
      break;
    };
    case WXK_INSERT:
      {
        if (!(m_mode & wxLC_SINGLE_SEL))
	{
          m_current->DrawRubberBand( dc, FALSE );
	  UnfocusLine( m_current );
          m_current->ReverseHilight( dc );
          wxNode *node = m_lines.Member( m_current )->Next();       
          if (node) m_current = (wxRawListLine*)node->Data();
          MoveToFocus();
          m_current->DrawRubberBand( dc, TRUE );
	  FocusLine( m_current );
	};
      };
      break;
    case WXK_RETURN:
    case WXK_EXECUTE:
      {
        ActivateLine( m_current );
      };
      break;
   };
  dc->SetBrush( oldBrush );
  dc->SetPen( oldPen );
};

void wxRawListCtrl::OnSetFocus( void )
{
};

void wxRawListCtrl::OnKillFocus( void )
{
};

void wxRawListCtrl::SetSize( int x, int y, int width, int height, int sizeFlags )
{
  m_dirty = TRUE;
  if (m_mode & wxLC_REPORT)
  {
    wxCanvas::SetSize( x, y+30, width, height-30, sizeFlags );
    m_titlesWindow->SetSize( x+3, y+2, width-4, 30 );
    m_titlesWindow->Show( TRUE );
  }
  else
  {
    wxCanvas::SetSize( x, y, width, height, sizeFlags );
    m_titlesWindow->Show( FALSE );
  };   
  Refresh();
};

wxFont *wxRawListCtrl::GetMyFont( void )
{
  return m_myFont;
};

void wxRawListCtrl::DrawImage( int index, wxCanvasDC *dc, int x, int y )
{
  if ((m_mode & wxLC_ICON) && (m_normal_image_list))
  {
    m_normal_image_list->Draw( index, *dc, x, y );
    return;
  };
  if ((m_mode & wxLC_SMALL_ICON) && (m_small_image_list))
  {
    m_small_image_list->Draw( index, *dc, x, y );
  };
};

void wxRawListCtrl::GetImageSize( int index, int &width, int &height )
{
  if ((m_mode & wxLC_ICON) && (m_normal_image_list))
  {
    m_normal_image_list->GetSize( index, width, height );
    return;
  };
  if ((m_mode & wxLC_SMALL_ICON) && (m_small_image_list))
  {
    m_small_image_list->GetSize( index, width, height );
    return;
  };
  width = 0;
  height = 0;
};

int wxRawListCtrl::GetTextLength( wxString &s )
{
  wxCanvasDC *dc = GetDC();
  float fw = 0;
  float fh = 0;
  dc->GetTextExtent( s, &fw, &fh );
  return (int)fw + 6;
};

int wxRawListCtrl::GetIndexOfLine( const wxRawListLine *line )
{
  int i = 0;
  wxNode *node = m_lines.First();
  while (node)
  {
    if (line == (wxRawListLine*)node->Data()) return i;
    i++;
    node = node->Next();
  };
  return -1;
};

void wxRawListCtrl::RawSetImageList( wxImageList *imageList, const int which )
{
  m_dirty = TRUE;
  if (which == wxIMAGE_LIST_NORMAL) m_normal_image_list = imageList;
  if (which == wxIMAGE_LIST_SMALL) m_small_image_list = imageList;
};

void wxRawListCtrl::RawSetItemSpacing( const int spacing, const Bool isSmall )
{
  m_dirty = TRUE;
  if (isSmall)
  { 
    m_small_spacing = spacing;
  }
  else
  {
    m_normal_spacing = spacing;
  };
};

int wxRawListCtrl::RawGetItemSpacing( const Bool isSmall )
{
  if (isSmall) return m_small_spacing; else return m_normal_spacing;
};

void wxRawListCtrl::RawSetColumn( const int col, wxListItem &item )
{
  m_dirty = TRUE;
  wxNode *node = m_columns.Nth( col );
  if (node)
  {
    if (item.m_width == wxLIST_AUTOSIZE_USEHEADER) item.m_width = GetTextLength( item.m_text )+7;
    wxRawListColumnInfo *column = (wxRawListColumnInfo*)node->Data();
    column->SetItem( item );
  };
};

void wxRawListCtrl::RawSetColumnWidth( const int col, const int width )
{
  m_dirty = TRUE;
  wxNode *node = m_columns.Nth( col );
  if (node)
  {
    wxRawListColumnInfo *column = (wxRawListColumnInfo*)node->Data();
    column->SetWidth( width );
  };
};

void wxRawListCtrl::RawGetColumn( const int col, wxListItem &item )
{
  wxNode *node = m_columns.Nth( col );
  if (node)
  {
    wxRawListColumnInfo *column = (wxRawListColumnInfo*)node->Data();
    column->GetItem( item );
  }
  else
  {
    item.m_format = 0;
    item.m_width = 0;
    item.m_text = "";
    item.m_image = 0;
    item.m_data = 0;
  };
};

int wxRawListCtrl::RawGetColumnWidth( const int col ) 
{
  wxNode *node = m_columns.Nth( col );
  if (node)
  {
    wxRawListColumnInfo *column = (wxRawListColumnInfo*)node->Data();
    return column->GetWidth();
  }
  else
    return 0;
};

int wxRawListCtrl::RawGetColumnCount( void )
{
  return m_columns.Number();
};

int wxRawListCtrl::RawGetCountPerPage( void )
{
  return m_visibleLines;
};

void wxRawListCtrl::RawSetItem( wxListItem &item )
{
  m_dirty = TRUE;
  wxNode *node = m_lines.Nth( item.m_itemId );
  if (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    if (m_mode & wxLC_REPORT) item.m_width = RawGetColumnWidth( item.m_col )-3;
    line->SetItem( item.m_col, item );
  };
};

void wxRawListCtrl::RawSetItemState( const long item, const long state, const long stateMask )
{
  // m_dirty = TRUE; no recalcs needed
  if (stateMask & wxLIST_STATE_FOCUSED) 
  {
    wxNode *node = m_lines.Nth( item );
    if (node) 
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      m_current->DrawRubberBand( dc, FALSE ); 
      UnfocusLine( m_current );
      m_current = line;
      m_current->DrawRubberBand( dc, TRUE );
      FocusLine( m_current );
    };
  };
  if (stateMask & wxLIST_STATE_SELECTED)
  {
    wxNode *node = m_lines.Nth( item );
    if (node) 
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      Bool on = state & wxLIST_STATE_SELECTED;
      line->DrawHilight( dc, on );
    };
  };
};

int wxRawListCtrl::RawGetItemState( const long item, const long stateMask )
{
  int ret = wxLIST_STATE_DONTCARE;
  if (stateMask & wxLIST_STATE_FOCUSED)
  {
    wxNode *node = m_lines.Nth( item );
    if (node) 
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      if (line == m_current) ret |= wxLIST_STATE_FOCUSED;
    };
  };
  if (stateMask & wxLIST_STATE_SELECTED)
  {
    wxNode *node = m_lines.Nth( item );
    if (node) 
    {
      wxRawListLine *line = (wxRawListLine*)node->Data();
      if (line->IsHilighted()) ret |= wxLIST_STATE_FOCUSED;
    };
  };
  return ret;
};

void wxRawListCtrl::RawGetItem( wxListItem &item )
{
  wxNode *node = m_lines.Nth( item.m_itemId );
  if (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    line->GetItem( item.m_col, item );
  }
  else
  {
    item.m_mask = 0;
    item.m_text = "";
    item.m_image = 0;
    item.m_data = 0;
  };
};

int wxRawListCtrl::RawGetItemCount( void )
{
  return m_lines.Number();
};

void wxRawListCtrl::RawGetItemRect( const long index, wxRectangle &rect )
{
  wxNode *node = m_lines.Nth( index );
  if (node) 
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    line->GetRect( rect );
  }
  else
  {
    rect.x = 0;
    rect.y = 0;
    rect.width = 0;
    rect.height = 0;
  };
};

int wxRawListCtrl::RawGetSelectedItemCount( void )
{
  int ret = 0;
  wxNode *node = m_lines.First();
  while (node)
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    if (line->IsHilighted()) ret++;
    node = node->Next();
  };
  return 0;
};

void wxRawListCtrl::RawSetMode( const long mode )
{
  m_dirty = TRUE;
  RawDeleteEverything();
  int x = 0;
  int y = 0;
  int width = 0;
  int height = 0;
  GetPosition( &x, &y );
  GetSize( &width, &height );

  if (mode & wxLC_REPORT)
  {
    if (!(m_mode & wxLC_REPORT))
    {
      wxCanvas::SetSize( x, y+30, width, height-30 );
      m_titlesWindow->SetSize( x+3, y+2, width-4, 30 );
      m_titlesWindow->Show( TRUE );
    };
  }
  else
  {
    if (m_mode & wxLC_REPORT)
    {
      wxCanvas::SetSize( x, y-30, width, height+30 );
      Refresh();
      m_titlesWindow->Show( FALSE );
    };
  };   
  m_mode = mode;
  if (m_mode & wxLC_REPORT)
  {
    m_xScroll = 0;
    m_yScroll = 5;
    if (GetScrollPos( wxHORIZONTAL ) != 0) SetScrollPos( wxHORIZONTAL, 0);
    SetScrollbars( m_xScroll, m_yScroll, 1, 0, 0, 20 );
  }
  else
  {
    m_xScroll = 5;
    m_yScroll = 0;
    if (GetScrollPos( wxVERTICAL ) != 0) SetScrollPos( wxVERTICAL, 0);
    SetScrollbars( m_xScroll, m_yScroll, 1, 0, 20, 0 );
  };
  MoveToFocus();
};

long wxRawListCtrl::RawGetMode( void ) const
{
  return m_mode;
};

void wxRawListCtrl::RawRealizeChanges( void )
{
  if (!m_current)
  {
    wxNode *node = m_lines.First();
    if (node) m_current = (wxRawListLine*)node->Data();
  };
  if (m_current)
  { 
    FocusLine( m_current );
    if (m_mode & wxLC_SINGLE_SEL) m_current->DrawHilight( GetDC(), TRUE );
  };
  Refresh();
};

long wxRawListCtrl::RawGetNextItem( const long item, int WXUNUSED(geometry), int state )
{
  long ret = 0;
  if (item > 0) ret = item;
  wxNode *node = m_lines.Nth( ret );
  while (node)
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    if ((state & wxLIST_STATE_FOCUSED) && (line == m_current)) return ret;
    if ((state & wxLIST_STATE_SELECTED) && (line->IsHilighted())) return ret;
    if (!state) return ret;
    ret++;
    node = node->Next();
  };
  return -1;
};

void wxRawListCtrl::RawDeleteItem( const long index )
{
  m_dirty = TRUE;
  wxNode *node = m_lines.Nth( index );
  if (node)
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    DeleteLine( line );
    m_lines.DeleteNode( node );
  };
};

void wxRawListCtrl::RawDeleteColumn( const int col )
{
  m_dirty = TRUE;
  wxNode *node = m_columns.Nth( col );
  if (node) m_columns.DeleteNode( node );
};

void wxRawListCtrl::RawDeleteEverything( void )
{
  m_dirty = TRUE;
  wxNode *node = m_lines.First();
  while (node)
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    DeleteLine( line );
    node = node->Next();
  };
  m_lines.Clear();
  m_current = NULL;
  m_columns.Clear();
};

void wxRawListCtrl::RawEnsureVisible( const long index )
{
  wxRawListLine *oldCurrent = m_current;
  m_current = NULL;
  int i = index;
  wxNode *node = m_lines.Nth( i );
  if (node) m_current = (wxRawListLine*)node->Data();
  if (m_current) MoveToFocus();
  m_current = oldCurrent;
};

long wxRawListCtrl::RawHitTest( const int x, const int y, int &flags )
{
  wxNode *node = m_lines.First();
  int count = 0;
  while (node)
  {
    wxRawListLine *line = (wxRawListLine*)node->Data();
    long ret = line->IsHit( x, y );
    if (ret & flags)
    {
      flags = ret;
      return count;
    };
    node = node->Next();
    count++;
  };
  return -1;
};

void wxRawListCtrl::RawInsertItem( wxListItem &item )
{
  m_dirty = TRUE;
  int mode = 0;
  if (m_mode & wxLC_REPORT) mode = wxLC_REPORT;
  else if (m_mode & wxLC_LIST) mode = wxLC_LIST;
  else if (m_mode & wxLC_ICON) mode = wxLC_ICON;
  else if (m_mode & wxLC_SMALL_ICON) mode = wxLC_ICON;  // no typo
  wxRawListLine *line = new wxRawListLine( this, mode, m_hilightBrush );
  if (m_mode & wxLC_REPORT)
  {
    line->InitItems( RawGetColumnCount() );
    item.m_width = RawGetColumnWidth( 0 )-3;
  }
  else 
    line->InitItems( 1 );
  line->SetItem( 0, item );
  wxNode *node = m_lines.Nth( item.m_itemId );
  if (node)
    m_lines.Insert( node, line );
  else
    m_lines.Append( line );
};

void wxRawListCtrl::RawInsertColumn( const long col, wxListItem &item )
{
  m_dirty = TRUE;
  if (m_mode & wxLC_REPORT)
  {
    if (item.m_width == wxLIST_AUTOSIZE_USEHEADER) item.m_width = GetTextLength( item.m_text );
    wxRawListColumnInfo *column = new wxRawListColumnInfo( item );
    wxNode *node = m_columns.Nth( col );
    if (node)
      m_columns.Insert( node, column );
    else
      m_columns.Append( column );
  };
};

wxListCtrlCompare list_ctrl_compare_func_2;
long              list_ctrl_compare_data;

int list_ctrl_compare_func_1( const void *arg1, const void *arg2 )
{
  wxRawListLine *line1 = *((wxRawListLine**)arg1);
  wxRawListLine *line2 = *((wxRawListLine**)arg2);
  wxListItem item;
  line1->GetItem( 0, item );
  long data1 = item.m_data;
  line2->GetItem( 0, item );
  long data2 = item.m_data;
  return list_ctrl_compare_func_2( data1, data2, list_ctrl_compare_data );
};

void wxRawListCtrl::RawSortItems( wxListCtrlCompare fn, long data )
{
  list_ctrl_compare_func_2 = fn;
  list_ctrl_compare_data = data;
  m_lines.Sort( list_ctrl_compare_func_1 );
};

Bool wxRawListCtrl::OnListNotify( wxListEvent &WXUNUSED(event) )
{ 
  return FALSE;
};


