/*
 * Copyright: (C) 1997, GNU (Robert Roebling)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef RawListCtrl_h
#define RawListCtrl_h

#ifdef __GNUG__
#pragma interface
#endif

#include "WIN_dnd.h"

//-----------------------------------------------------------------------------
// classes
//-----------------------------------------------------------------------------

class wxRawListItemInfo;
class wxRawListColumnInfo;
class wxRawListLine;
class ColumnTitles;
class wxRawListCtrl;

//-----------------------------------------------------------------------------
//  constants and types
//-----------------------------------------------------------------------------

enum viewType  { wxVIEW_SHORT, wxVIEW_TABLE };

//-----------------------------------------------------------------------------
//  wxRawListItemInfo
//-----------------------------------------------------------------------------

class wxRawListItemInfo : public wxObject
{
  DECLARE_DYNAMIC_CLASS(wxRawListItemInfo);

  protected:
    wxString   m_text;
    int        m_image;
    long       m_data;
    int        m_xpos,m_ypos;
    int        m_width,m_height;
    wxColour   *m_colour;
    
  public:
    wxRawListItemInfo(void);
    wxRawListItemInfo( const wxListItem &info );
    void SetItem( const wxListItem &info );
    void SetText( const wxString &s );
    void SetImage( const int image );
    void SetData( const long data );
    void SetPosition( const int x, const int y );
    void SetSize( const int width, const int height );
    void SetColour( wxColour *col );
    Bool HasImage(void) const;
    Bool HasText(void) const;
    Bool IsHit( const int x, const int y ) const;
    void GetText( wxString &s );
    int GetX( void ) const;
    int GetY( void ) const;
    int GetWidth(void) const;
    int GetHeight(void) const;
    int GetImage(void) const;
    void GetItem( wxListItem &info );
    wxColour *GetColour(void);
};

//-----------------------------------------------------------------------------
//  wxRawListColumnInfo
//-----------------------------------------------------------------------------

class wxRawListColumnInfo : public wxObject
{
  DECLARE_DYNAMIC_CLASS(wxRawListColumnInfo);

  protected:
    long      m_mask;
    int       m_image;
    wxString  m_text;
    int       m_format;
    int       m_width;
    int       m_xpos,m_ypos;
    int       m_height;
    
  public:
    wxRawListColumnInfo(void);
    wxRawListColumnInfo( const wxListItem &info );
    void SetItem( const wxListItem &item );
    void SetPosition( const int x, const int y );
    void SetWidth( const int w );
    void SetFormat( const int format );
    void SetHeight( const int h );
    Bool HasImage(void) const;
    Bool HasText(void) const;
    Bool IsHit( const int x, const int y ) const;
    void GetItem( wxListItem &item );
    void GetText( wxString &s );
    int GetImage(void) const;
    int GetWidth(void) const;
    int GetFormat(void) const;
};

//-----------------------------------------------------------------------------
//  wxRawListLine
//-----------------------------------------------------------------------------

class wxRawListLine : public wxObject
{
  DECLARE_DYNAMIC_CLASS(wxRawListLine);

  protected:
    wxList         m_items;
    wxRectangle    m_bound_all;
    wxRectangle    m_bound_label;
    wxRectangle    m_bound_icon;
    wxRectangle    m_bound_hilight;
    int            m_mode;
    Bool           m_hilighted;
    wxRawListCtrl *m_owner;
    wxBrush       *m_hilightBrush;
    int            m_spacing;

    void DoDraw( wxCanvasDC *dc, const Bool hilight, const Bool paintBG );

  public:
    wxRawListLine( void ) {};
    wxRawListLine( wxRawListCtrl *owner, const int mode, wxBrush *hilightBrush );
    void CalculateSize( wxCanvasDC *dc, const int spacing );
    void SetPosition( wxCanvasDC *dc, const int x, const int y,  const int window_width );
    void SetColumnPosition( const int index, const int x );
    void GetSize( int &width, int &height );
    void GetExtent( int &x, int &y, int &width, int &height );
    long IsHit( const int x, const int y );
    void InitItems( const int num );
    void SetItem( const int index, const wxListItem &info );
    void GetItem( const int index, wxListItem &info );
    void GetText( const int index, wxString &s );
    void SetText( const int index, const wxString s );
    int GetImage( const int index );
    void GetRect( wxRectangle &rect );
    void DrawHilight( wxCanvasDC *dc, const Bool on );
    void ReverseHilight( wxCanvasDC *dc );
    void DrawRubberBand( wxCanvasDC *dc, const Bool on );
    void Draw( wxCanvasDC *dc );
    Bool IsInRect( const int x, const int y, const wxRectangle &rect );
    Bool IsHilighted( void );
    void AssignRect( wxRectangle &dest, const int x, const int y, const int width, const int height );
    void AssignRect( wxRectangle &dest, const wxRectangle &source );
};

//-----------------------------------------------------------------------------
//  wxColumnTitles
//-----------------------------------------------------------------------------

class wxColumnTitles : public wxPanel
{
  DECLARE_DYNAMIC_CLASS(wxColumnTitles)

  protected:
    wxRawListCtrl  *m_owner;
    wxCursor       *m_currentCursor;
    wxCursor       *m_resizeCursor;

  public:    
    wxColumnTitles( void ) : wxPanel() {};
    wxColumnTitles( wxWindow *win, wxRawListCtrl *owner, int x, int y, int width, int height, 
      long style=0, Constdata char *name= "wxColumnTitles" );
    void DoDrawRect( wxCanvasDC *dc, int x, int y, int w, int h );
    void DoDrawFrame( wxCanvasDC *dc, int x, int y, int w, int h );
    void OnPaint( void );
    void OnEvent( wxMouseEvent &event );
    void OnSetFocus( void );
};

//-----------------------------------------------------------------------------
//  wxRawListCtrl
//-----------------------------------------------------------------------------

class wxRawListCtrl : public wxDragCanvas
{
  DECLARE_DYNAMIC_CLASS(wxRawListCtrl);

  protected:
    long             m_mode;
    wxList           m_lines;
    wxList           m_columns;
    wxRawListLine   *m_current;
    int              m_visibleLines;
    wxBrush         *m_hilightBrush;
    wxColour        *m_hilightColour;
    wxFont          *m_myFont;
    wxColumnTitles  *m_titlesWindow;
    int              m_xScroll,m_yScroll;
    Bool             m_dirty;
    wxImageList     *m_small_image_list;
    wxImageList     *m_normal_image_list;
    int              m_small_spacing;
    int              m_normal_spacing;

  public:
    wxRawListCtrl( void ); 
    wxRawListCtrl( wxWindow *parent, int x, int y, int width, int height, 
      long mode, long style=0, Constdata char *name= "wxRawListCtrl" );
    ~wxRawListCtrl( void );
    void OnPaint( void );
    void CalculatePositions( void );
    void HilightAll( const Bool on );
    void ActivateLine( wxRawListLine *line );
    void SendNotify( wxRawListLine *line, long command );
    void FocusLine( wxRawListLine *line );
    void UnfocusLine( wxRawListLine *line );
    void SelectLine( wxRawListLine *line );
    void DeselectLine( wxRawListLine *line );
    void DeleteLine( wxRawListLine *line );
    void RenameLine( wxRawListLine *line, const wxString &newName );
    void OnEvent( wxMouseEvent &event );
    void MoveToFocus( void );
    void OnArrowChar( wxRawListLine *newCurrent, Bool shiftDown );
    void OnChar( wxKeyEvent &event );
    void OnSetFocus( void );
    void OnKillFocus( void );
    void SetSize( int x, int y, int width, int height, int sizeFlags = wxSIZE_AUTO );
    wxFont *GetMyFont( void );
    void DrawImage( int index, wxCanvasDC *dc, int x, int y );
    void GetImageSize( int index, int &width, int &height );
    int GetIndexOfLine( const wxRawListLine *line );
    int GetTextLength( wxString &s );  // should be const

    void RawSetImageList( wxImageList *imageList, const int which );
    void RawSetItemSpacing( const int spacing, const Bool isSmall = FALSE );
    int RawGetItemSpacing( const Bool isSmall = FALSE );
    void RawSetColumn( const int col, wxListItem &item );
    void RawSetColumnWidth( const int col, const int width );
    void RawGetColumn( const int col, wxListItem &item );
    int RawGetColumnWidth( const int vol );
    int RawGetColumnCount( void );
    int RawGetCountPerPage( void );     
    void RawSetItem( wxListItem &item );
    void RawGetItem( wxListItem &item );
    void RawSetItemState( const long item, const long state, const long stateMask ); 
    int RawGetItemState( const long item, const long stateMask );
    int RawGetItemCount( void );
    void RawGetItemRect( const long index, wxRectangle &rect );
    int RawGetSelectedItemCount( void );
    void RawSetMode( const long mode );
    long RawGetMode( void ) const;
    void RawRealizeChanges( void );
    long RawGetNextItem( const long item, int geometry, int state );
    void RawDeleteItem( const long index );
    void RawDeleteColumn( const int col );
    void RawDeleteEverything( void );
    void RawEnsureVisible( const long index );
    long RawHitTest( const int x, const int y, int &flags );
    void RawInsertItem( wxListItem &item );
    void RawInsertColumn( const long col, wxListItem &item );
    void RawSortItems( wxListCtrlCompare fn, long data );
    virtual Bool OnListNotify( wxListEvent &event );
};

#endif // RawListCtrl_h
